package ru.dbotthepony.mc.otm.block.entity.decorative

import com.google.common.collect.ImmutableList
import it.unimi.dsi.fastutil.ints.Int2IntArrayMap
import net.minecraft.core.BlockPos
import net.minecraft.nbt.CompoundTag
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.DyeColor
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.container.HandlerFilter
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.core.immutableList
import ru.dbotthepony.mc.otm.core.immutableMap
import ru.dbotthepony.mc.otm.core.nbt.mapPresent
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.menu.decorative.PainterMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import java.util.*

class PainterBlockEntity(blockPos: BlockPos, blockState: BlockState) : MatteryDeviceBlockEntity(MBlockEntities.PAINTER, blockPos, blockState) {
	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return PainterMenu(containerID, inventory, this)
	}

	val dyeInput = MatteryContainer(this::markDirtyFast, 1)
	private val dyeStored = EnumMap<DyeColor, Int>(DyeColor::class.java)
	val dyeStoredView: Map<DyeColor, Int> = Collections.unmodifiableMap(dyeStored)

	init {
		addDroppableContainer(dyeInput)
		savetables.stateful(dyeInput, INVENTORY_KEY)
	}

	fun takeDyes(dyes: Map<DyeColor, Int>) {
		for ((dye, amount) in dyes) {
			for (i in 0 until amount) {
				mixer(dye).mix(dyeStored)

				if (dyeStored(dye) > 0) {
					dyeStored[dye] = dyeStored(dye) - 1
				}
			}
		}

		markDirtyFast()
	}

	val config = ConfigurableItemHandler(input = dyeInput.handler(object : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			val dye = DyeColor.entries.firstOrNull { stack.`is`(it.tag) } ?: return false
			return dyeStored(dye) + HUE_PER_ITEM <= MAX_STORAGE
		}

		override fun modifyInsertCount(slot: Int, stack: ItemStack, existing: ItemStack, simulate: Boolean): Int {
			if (!stack.equals(existing, false))
				return super.modifyInsertCount(slot, stack, existing, simulate)

			val dye = DyeColor.entries.firstOrNull { stack.`is`(it.tag) } ?: return 0
			return stack.count.coerceAtMost((MAX_STORAGE - dyeStored(dye)) / HUE_PER_ITEM - existing.count)
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return false
		}
	}))

	fun dyeStored(dye: DyeColor): Int {
		return dyeStored[dye]!!
	}

	init {
		for (dye in DyeColor.entries) {
			dyeStored[dye] = 0
		}
	}

	override fun saveShared(nbt: CompoundTag) {
		super.saveShared(nbt)

		nbt["dyes"] = CompoundTag().also {
			for ((k, v) in dyeStored) {
				it[k.getName()] = v
			}
		}
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		for (dye in DyeColor.entries) {
			dyeStored[dye] = 0
		}

		nbt.mapPresent("dyes") { it: CompoundTag ->
			for (k in it.allKeys) {
				dyeStored[DyeColor.entries.firstOrNull { it.getName() == k } ?: continue] = it.getInt(k)
			}
		}
	}

	override fun tick() {
		super.tick()

		val i = dyeInput.iterator()

		for (item in i) {
			val dye = DyeColor.entries.firstOrNull { item.`is`(it.tag) } ?: continue
			val stored = dyeStored(dye)

			if (stored + HUE_PER_ITEM <= MAX_STORAGE) {
				item.shrink(1)
				i.setChanged()
				dyeStored[dye] = stored + HUE_PER_ITEM
			}
		}
	}

	class Mixer(val color: DyeColor, vararg mixing: ImmutableList<DyeColor>) : Map.Entry<DyeColor, Mixer> {
		val mixing: ImmutableList<ImmutableList<DyeColor>> = ImmutableList.copyOf(mixing)

		override val key: DyeColor
			get() = color
		override val value: Mixer
			get() = this

		private fun mix(input: MutableMap<DyeColor, Int>, seen: MutableSet<DyeColor>, stack: MutableSet<DyeColor>) {
			if (input[color]!! > 0 || color in seen || color in stack) return
			stack.add(color)

			for (ingredients in mixing) {
				val copy = EnumMap(input)
				var i = 0
				val volumes = Int2IntArrayMap()

				// страховка от взаимной блокировки когда последующий ингредиент требует предыдущий.
				// создаёт горку из значений
				for (i2 in ingredients.indices)
					volumes[i2] = 1

				while (!ingredients.all { copy[it]!! > 0 } && i++ < 20) {
					ingredients.withIndex().forEach { for (t in 0 until volumes[it.index]) mixer(it.value).mix(copy, seen, stack) }

					var increase = i
					val iter = ingredients.indices.iterator()

					while (iter.hasNext() && increase-- > 0) {
						volumes[iter.nextInt()]++
					}
				}

				if (ingredients.all { copy[it]!! > 0 }) {
					ingredients.forEach { copy[it] = copy[it]!! - 1 }
					copy[color] = ingredients.size

					input.putAll(copy)
					stack.remove(color)
					return
				}
			}

			stack.remove(color)
			seen.add(color)
		}

		fun mix(input: MutableMap<DyeColor, Int>) {
			mix(input, EnumSet.noneOf(DyeColor::class.java), EnumSet.noneOf(DyeColor::class.java))
		}

		fun isAvailable(input: Map<DyeColor, Int>): Boolean {
			if (input[color]!! > 0) return true
			return EnumMap(input).also(::mix)[color]!! > 0
		}
	}

	companion object {
		fun mixer(color: DyeColor): Mixer {
			return MIXING[color]!!
		}

		val MIXING = immutableMap {
			put(Mixer(DyeColor.WHITE, immutableList(DyeColor.RED, DyeColor.GREEN, DyeColor.BLUE)))
			put(Mixer(DyeColor.LIGHT_GRAY, immutableList(DyeColor.BLACK, DyeColor.WHITE, DyeColor.WHITE)))
			put(Mixer(DyeColor.GRAY, immutableList(DyeColor.BLACK, DyeColor.WHITE)))
			put(Mixer(DyeColor.ORANGE, immutableList(DyeColor.YELLOW, DyeColor.RED)))
			put(Mixer(DyeColor.LIME, immutableList(DyeColor.GREEN, DyeColor.WHITE)))
			put(Mixer(DyeColor.LIGHT_BLUE, immutableList(DyeColor.BLUE, DyeColor.WHITE)))
			put(Mixer(DyeColor.CYAN, immutableList(DyeColor.BLUE, DyeColor.GREEN)))
			put(Mixer(DyeColor.PURPLE, immutableList(DyeColor.BLUE, DyeColor.RED)))

			put(Mixer(
				DyeColor.MAGENTA,
				immutableList(DyeColor.PURPLE, DyeColor.PINK),
				immutableList(DyeColor.RED, DyeColor.RED, DyeColor.BLUE, DyeColor.WHITE),
				immutableList(DyeColor.PINK, DyeColor.RED, DyeColor.BLUE)
			))

			put(Mixer(DyeColor.PINK, immutableList(DyeColor.RED, DyeColor.WHITE)))

			put(Mixer(DyeColor.RED, immutableList(DyeColor.MAGENTA, DyeColor.YELLOW)))
			put(Mixer(DyeColor.GREEN, immutableList(DyeColor.CYAN, DyeColor.YELLOW)))
			put(Mixer(DyeColor.BLUE, immutableList(DyeColor.CYAN, DyeColor.MAGENTA)))
			put(Mixer(DyeColor.BLACK, immutableList(DyeColor.CYAN, DyeColor.MAGENTA, DyeColor.YELLOW)))
			put(Mixer(DyeColor.YELLOW, immutableList(DyeColor.RED, DyeColor.GREEN)))
			put(Mixer(DyeColor.BROWN, immutableList(DyeColor.MAGENTA, DyeColor.YELLOW, DyeColor.BLACK)))
		}

		const val MAX_STORAGE = 256
		const val HUE_PER_ITEM = 32
	}
}
