package ru.dbotthepony.mc.otm.core.collect

import it.unimi.dsi.fastutil.objects.ObjectIterators
import ru.dbotthepony.mc.otm.core.addAll
import java.util.Optional
import java.util.Spliterator
import java.util.Spliterators
import java.util.function.BinaryOperator
import java.util.function.Predicate
import java.util.function.Supplier
import java.util.stream.Collector
import java.util.stream.Stream
import java.util.stream.StreamSupport

// Purpose of Stream API over Iterators is that it is simple enough for JIT to inline most of it,
// unlike actual Streams.

// We lose only one (actual) element of Stream API here: ability to parallelize work,
// except it doesn't (properly) work in Minecraft (Forge) either, due to classloader bug:
// https://github.com/MinecraftForge/EventBus/issues/44

// Aside parallel work, unimplemented Stream API elements can be easily implemented when required

/**
 * Filters elements of [parent] iterator
 *
 * Resulting [Iterator] is [MutableIterator] if [parent] is
 */
class FilteringIterator<T>(private val parent: Iterator<T>, private val predicate: Predicate<in T>) : MutableIterator<T> {
	private var foundValue: Any? = Companion
	private var returned = false

	override fun hasNext(): Boolean {
		if (foundValue === Companion) {
			while (parent.hasNext()) {
				val next = parent.next()

				if (predicate.test(next)) {
					foundValue = next
					return true
				}
			}

			return false
		}

		return true
	}

	@Suppress("unchecked_cast")
	override fun next(): T {
		var foundValue = foundValue

		if (foundValue === Companion) {
			if (!hasNext()) {
				throw NoSuchElementException()
			}

			foundValue = this.foundValue

			if (foundValue === Companion) {
				throw ConcurrentModificationException()
			}
		}

		this.foundValue = Companion
		returned = true
		return foundValue as T
	}

	override fun remove() {
		if (!returned) throw NoSuchElementException()
		(parent as MutableIterator<T>).remove()
		returned = false
	}

	private companion object
}

/**
 * Maps elements of [parent] iterator from values of [T] to [R] using function [mapper]
 *
 * Resulting [Iterator] is [MutableIterator] if [parent] is
 */
class MappingIterator<T, R>(private val parent: Iterator<T>, private val mapper: (T) -> R) : MutableIterator<R> {
	override fun hasNext(): Boolean {
		return parent.hasNext()
	}

	override fun next(): R {
		return mapper.invoke(parent.next())
	}

	override fun remove() {
		(parent as MutableIterator<T>).remove()
	}
}

/**
 * Maps elements of [parent] iterator from type [T] to other iterators of type [R] using function [mapper]
 *
 * Resulting [Iterator] is [MutableIterator] if [parent] is
 */
class FlatMappingIterator<T, R>(private val parent: Iterator<T>, private val mapper: (T) -> Iterator<R>) : MutableIterator<R> {
	private var current: Iterator<R>? = null
	private var last: Iterator<R>? = null

	override fun hasNext(): Boolean {
		while (current?.hasNext() != true && parent.hasNext()) {
			current = mapper.invoke(parent.next())
		}

		return current?.hasNext() == true
	}

	override fun next(): R {
		if (!hasNext())
			throw NoSuchElementException()

		val v = current!!.next()
		last = current
		return v
	}

	override fun remove() {
		(last as MutableIterator<R>? ?: throw NoSuchElementException()).remove()
		last = null
	}
}

/**
 * Limits amount of values returned by [parent] iterator to return at most [limit] values
 *
 * Resulting [Iterator] is [MutableIterator] if [parent] is
 */
class LimitingIterator<T>(private val parent: Iterator<T>, private val limit: Long) : Iterator<T> {
	init {
		require(limit > 0) { "Invalid limit $limit" }
	}

	private var found = 0L

	override fun hasNext(): Boolean {
		return found < limit && parent.hasNext()
	}

	override fun next(): T {
		if (found >= limit)
			throw NoSuchElementException()

		return parent.next()
	}
}

/**
 * Skips (discards) up to [skip] values returned by [parent] iterator
 *
 * Resulting [Iterator] is [MutableIterator] if [parent] is
 */
class SkippingIterator<T>(private val parent: Iterator<T>, skip: Long) : MutableIterator<T> {
	init {
		require(skip >= 0) { "Invalid skip amount $skip" }
	}

	private var found = skip
	private var returned = false

	override fun hasNext(): Boolean {
		while (parent.hasNext() && found > 0L) {
			found--
			parent.next()
		}

		return parent.hasNext()
	}

	override fun next(): T {
		if (!hasNext())
			throw NoSuchElementException()

		val v = parent.next()
		returned = true
		return v
	}

	override fun remove() {
		if (!returned) {
			throw NoSuchElementException()
		}

		returned = false
		(parent as MutableIterator<T>).remove()
	}
}

fun <T> concatIterators(): MutableIterator<T> {
	return ObjectIterators.EMPTY_ITERATOR as MutableIterator<T>
}

fun <T> concatIterators(a: Iterator<T>): MutableIterator<T> {
	return a as MutableIterator<T>
}

fun <T> concatIterators(iterators: Iterable<Iterator<T>>): MutableIterator<T> {
	return iterators.iterator().flatMap { it }
}

fun <T> concatIterators(vararg iterators: Iterator<T>): MutableIterator<T> {
	return iterators.iterator().flatMap { it }
}

/**
 * Filters elements of [this] iterator
 *
 * Resulting [Iterator] is [MutableIterator] if [this] is
 */
fun <T> Iterator<T>.filter(condition: Predicate<in T>) = FilteringIterator(this, condition)

/**
 * Maps elements of [this] iterator from values of [T] to [R] using function [mapper]
 *
 * Resulting [Iterator] is [MutableIterator] if [this] is
 */
fun <T, R> Iterator<T>.map(mapper: (T) -> R) = MappingIterator(this, mapper)

/**
 * Maps elements of [this] iterator from values of [T] to [R] using function [mapper]
 *
 * Resulting [Iterator] is [MutableIterator] if [this] is
 */
fun <T, R> Iterator<T>.map(mapper: java.util.function.Function<T, R>) = MappingIterator(this, mapper::apply)

/**
 * Maps elements of [this] iterator from type [T] to other iterators of type [R] using function [mapper]
 *
 * Resulting [Iterator] is [MutableIterator] if [this] is
 */
fun <T, R> Iterator<T>.flatMap(mapper: (T) -> Iterator<R>) = FlatMappingIterator(this, mapper)

/**
 * Maps elements of [this] iterator from type [T] to other iterators of type [R] using function [mapper]
 *
 * Resulting [Iterator] is [MutableIterator] if [this] is
 */
fun <T, R> Iterator<T>.flatMap(mapper: java.util.function.Function<T, Iterator<R>>) = FlatMappingIterator(this, mapper::apply)

fun <T> Iterator<T>.reduce(identity: T, reducer: (T, T) -> T): T {
	var result = identity
	for (value in this) result = reducer.invoke(result, value)
	return result
}

fun <T> Iterator<T>.reduce(identity: T, reducer: BinaryOperator<T>): T = reduce(identity, reducer::apply)
fun <T> Iterator<T?>.filterNotNull(): MutableIterator<T> = filter { it != null } as MutableIterator<T>

fun <T> Iterator<T>.any(predicate: Predicate<in T>): Boolean {
	for (value in this) {
		if (predicate.test(value)) {
			return true
		}
	}

	return false
}

fun <T> Iterator<T>.all(predicate: Predicate<in T>): Boolean {
	for (value in this) {
		if (!predicate.test(value)) {
			return false
		}
	}

	return true
}

fun <T> Iterator<T>.none(predicate: Predicate<in T>): Boolean {
	for (value in this) {
		if (predicate.test(value)) {
			return false
		}
	}

	return true
}

fun <T, A, R> Iterator<T>.collect(collector: Collector<T, A, R>): R {
	val accumulator = collector.accumulator()
	val instance = collector.supplier().get()

	for (value in this) {
		accumulator.accept(instance, value)
	}

	return collector.finisher().apply(instance)
}

fun <T> Iterator<T>.toList(): MutableList<T> {
	val result = ArrayList<T>()
	result.addAll(this)
	return result
}

fun <T : Any> Iterator<T>.find(): Optional<T> {
	if (hasNext()) {
		return Optional.of(next())
	}

	return Optional.empty<T>()
}

fun <T> Iterator<T>.limit(limit: Long) = LimitingIterator(this, limit)
fun <T> Iterator<T>.skip(skip: Long) = if (skip == 0L) this else SkippingIterator(this, skip)

inline fun <T> Iterator<T>.forEach(action: (T) -> Unit) {
	for (value in this) {
		action.invoke(value)
	}
}

fun <T : Any> Iterator<T>.min(comparator: Comparator<T>): Optional<T> {
	if (!hasNext()) {
		return Optional.empty<T>()
	}

	var min = next()

	for (value in this) {
		if (comparator.compare(min, value) > 0) {
			min = value
		}
	}

	return Optional.of(min)
}

fun <T : Any> Iterator<T>.max(comparator: Comparator<T>): Optional<T> {
	if (!hasNext()) {
		return Optional.empty<T>()
	}

	var max = next()

	for (value in this) {
		if (comparator.compare(max, value) < 0) {
			max = value
		}
	}

	return Optional.of(max)
}

fun <T> Iterator<T>.peek(peeker: (T) -> Unit): MutableIterator<T> {
	return object : MutableIterator<T> {
		override fun hasNext(): Boolean {
			return this@peek.hasNext()
		}

		override fun next(): T {
			return this@peek.next().also(peeker)
		}

		override fun remove() {
			(this@peek as MutableIterator<T>).remove()
		}
	}
}

fun <T> Iterator<T>.maybe(): T? {
	return if (hasNext())
		next()
	else
		null
}

fun <T> emptyIterator(): MutableIterator<T> {
	return ObjectIterators.emptyIterator()
}

fun <T> Iterator<T>.toStream(): Stream<T> {
	return StreamSupport.stream(Spliterators.spliteratorUnknownSize(this, 0), false)
}

fun <T> Iterator<T>.allEqual(): Boolean {
	if (hasNext()) {
		val v = next()

		while (hasNext()) {
			if (v != next()) {
				return false
			}
		}

		return true
	}

	return false
}

fun <T> Iterator<T>.count(): Long {
	var count = 0L
	while (hasNext()) count++
	return count
}
