package ru.dbotthepony.mc.otm.network

import java.util.function.Function
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.player.Player
import net.minecraftforge.event.TickEvent
import net.minecraftforge.event.TickEvent.ServerTickEvent
import net.minecraftforge.event.server.ServerStoppedEvent
import net.minecraftforge.event.server.ServerStoppingEvent
import net.minecraftforge.network.NetworkDirection
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.network.NetworkRegistry
import net.minecraftforge.network.PacketDistributor
import net.minecraftforge.network.simple.SimpleChannel
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.NULLABLE_MINECRAFT_SERVER
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import java.math.BigDecimal
import java.math.BigInteger
import java.util.Optional
import java.util.concurrent.CompletableFuture
import java.util.concurrent.ConcurrentLinkedDeque
import java.util.concurrent.locks.LockSupport
import java.util.function.BiConsumer
import java.util.function.Supplier
import kotlin.reflect.KClass

fun Supplier<NetworkEvent.Context>.enqueueWork(lambda: Runnable): CompletableFuture<Void> = get().enqueueWork(lambda)
fun Supplier<NetworkEvent.Context>.enqueueWork(lambda: () -> Unit): CompletableFuture<Void> = get().enqueueWork(lambda)
var Supplier<NetworkEvent.Context>.packetHandled: Boolean
	get() = get().packetHandled
	set(value) { get().packetHandled = value }

val Supplier<NetworkEvent.Context>.sender: ServerPlayer?
	get() = get().sender

interface MatteryPacket {
	fun write(buff: FriendlyByteBuf)
	fun play(context: Supplier<NetworkEvent.Context>)
}

abstract class MatteryNetworkChannel(val version: String, val name: String) {
	val channel: SimpleChannel = NetworkRegistry.newSimpleChannel(
		ResourceLocation(OverdriveThatMatters.MOD_ID, name),
		{ version },
		{ it == version },
		{ it == version },
	)

	fun sendToServer(packet: Any) = channel.sendToServer(packet)

	fun send(ply: Player, packet: Any) {
		if (ply is ServerPlayer) {
			queue.add(Task(channel, PacketDistributor.PLAYER.with { ply }, packet))
		}
	}

	fun sendNow(ply: Player, packet: Any) {
		if (ply is ServerPlayer) {
			channel.send(PacketDistributor.PLAYER.with { ply }, packet)
		}
	}

	fun sendTracking(entity: Entity, packet: Any) {
		if ((NULLABLE_MINECRAFT_SERVER?.playerCount ?: 0) <= 0) {
			return
		}

		queue.add(Task(channel, PacketDistributor.TRACKING_ENTITY.with { entity }, packet))
	}

	fun sendTrackingAndSelf(entity: Entity, packet: Any) {
		if ((NULLABLE_MINECRAFT_SERVER?.playerCount ?: 0) <= 0) {
			return
		}

		queue.add(Task(channel, PacketDistributor.TRACKING_ENTITY_AND_SELF.with { entity }, packet))
	}

	fun send(distributor: PacketDistributor.PacketTarget, packet: Any) {
		queue.add(Task(channel, distributor, packet))
	}

	fun sendNow(distributor: PacketDistributor.PacketTarget, packet: Any) {
		channel.send(distributor, packet)
	}

	private var nextNetworkPacketID = 0

	fun <T> add(
		packetClass: Class<T>,
		writer: BiConsumer<T, FriendlyByteBuf>,
		reader: Function<FriendlyByteBuf, T>,
		handler: BiConsumer<T, Supplier<NetworkEvent.Context>>,
		direction: NetworkDirection? = null
	) {
		if (nextNetworkPacketID >= 256) {
			throw IndexOutOfBoundsException("Network message ID overflow!")
		}

		@Suppress("INACCESSIBLE_TYPE")
		channel.registerMessage(nextNetworkPacketID++, packetClass, writer, reader, handler, if (direction == null) Optional.empty() else Optional.of(direction))
	}

	fun <T : MatteryPacket> add(
		packetClass: Class<T>,
		reader: Function<FriendlyByteBuf, T>,
		direction: NetworkDirection? = null
	) {
		add(packetClass, MatteryPacket::write, reader, MatteryPacket::play, direction)
	}

	fun <T : MatteryPacket> add(
		packetClass: KClass<T>,
		reader: Function<FriendlyByteBuf, T>,
		direction: NetworkDirection? = null
	) {
		add(packetClass.java, MatteryPacket::write, reader, MatteryPacket::play, direction)
	}

	private data class Task(val channel: SimpleChannel, val target: PacketDistributor.PacketTarget, val packet: Any)

	companion object {
		private val logger = LogManager.getLogger()
		private var thread: Thread? = null

		private val queue = ConcurrentLinkedDeque<Task>()

		@Volatile
		private var interrupt = false

		/**
		 * Reason behind separate thread is that Minecraft connection class does not allow to enqueue messages,
		 * nor does Forge. This leads to extensive calls to Netty Channel eventLoop().execute(), which is somehow resource
		 * heavy, as it calls wakeup on Windows, eventFdWrite on Linux, etc., when all we want to do is to enqueue packet.
		 *
		 * Concurrent safety - Minecraft Channel (send method) does not do any writes (unless player is unconnected, in which case it uses thread safe queue), hence it is thread safe.
		 * OTM packets are always dispatched in correct order to client.
		 * Netty eventLoop executor is thread safe by definition.
		 */
		private fun run() {
			while (!interrupt) {
				val task = queue.pollFirst()

				if (task == null) {
					LockSupport.park()
				} else {
					try {
						task.channel.send(task.target, task.packet)
					} catch(err: Throwable) {
						logger.error("Error executing network dispatcher task", err)
					}
				}
			}

			logger.debug("Overdrive That Matters Packet Dispatcher thread exited gracefully")
		}

		internal fun onServerPostTick() {
			LockSupport.unpark(thread)
		}

		internal fun onServerStarting() {
			interrupt = false
			check(thread?.isAlive != true) { "Already having network dispatcher thread, ServerStartingEvent was fired twice!" }
			thread = Thread(this::run, "Overdrive That Matters Network Dispatcher").also { it.isDaemon = true; it.start() }
		}

		internal fun onServerStopping() {
			interrupt = true
			LockSupport.unpark(thread)
		}

		internal fun onServerStopped() {
			interrupt = true
			LockSupport.unpark(thread)
		}
	}
}
